try:
    import trimesh
    TRIMESH = True
except ImportError:
    TRIMESH = False
try:
    from vedo import show, Box, Axes, trimesh2vedo, Lines, Sphere
    VEDO = True
except ImportError:
    VEDO = False
import numpy as np

try:
    import plotly.graph_objects as go
    PLOTLY = True
except ImportError:
    PLOTLY = False

from .version import __version__ as VERSION


class World:
    def __init__(
        self,
        bb_center,
        length,
        width,
        height
    ):
        self.bb_center = bb_center
        self.length = length
        self.width = width
        self.height = height
        self.world = Box(
            bb_center,
            length,
            width,
            height
        ).wireframe()


def plot(
    data,
    names=None,
    colors=None,
    lines=None,
    targets=None,
    arrows=None,
    text=None,
    boxes=None,
    points=None,
    interactive=True,
):
    """
    A vedo wrapper for quickly visualizing well trajectories for QAQC purposes.

    Parameters
    ----------
        data: a trimesh.Trimesh object or a list of trimesh.Trimesh
        objects or a trmiesh.scene object
        names: list of strings (default: None)
            A list of names, index aligned to the list of well meshes.
        colors: list of strings (default: None)
            A list of color or colors. If a single color is listed then this is
            applied to all meshes in data, otherwise the list of colors is
            indexed to the list of meshes.
    """
    assert all((VEDO, TRIMESH)), "ImportError: try pip install welleng[easy]"
    if isinstance(data, trimesh.scene.scene.Scene):
        meshes = [v for k, v in data.geometry.items()]
        if names is None:
            names = list(data.geometry.keys())

    # handle a single mesh being passed
    elif isinstance(data, trimesh.Trimesh):
        meshes = [data]

    else:
        meshes = data
        if names is not None:
            assert len(names) == len(data), \
                "Names must be length of meshes list else None"

    if colors is not None:
        if len(colors) == 1:
            colors = colors * len(meshes)
        else:
            assert len(colors) == len(meshes), \
                "Colors must be length of meshes list, 1 else None"

    if points is not None:
        points = [
            Sphere(p, r=30, c='grey')
            for p in points
        ]

    meshes_vedo = []
    for i, mesh in enumerate(meshes):
        if i == 0:
            vertices = np.array(mesh.vertices)
            start_locations = np.array([mesh.vertices[0]])
        else:
            vertices = np.concatenate(
                (vertices, np.array(mesh.vertices)),
                axis=0
            )
            start_locations = np.concatenate(
                (start_locations, np.array([mesh.vertices[0]])),
                axis=0
            )

        # convert to vedo mesh
        m_vedo = trimesh2vedo(mesh)
        if colors is not None:
            m_vedo.c(colors[i])
        if names is not None:
            m_vedo.name = names[i]
            m_vedo.flag()
        meshes_vedo.append(m_vedo)

    w = get_bb(vertices)

    axes = get_axes(w.world)

    # try and figure out a nice start camera position
    pos = w.bb_center
    vec1 = pos - [w.length, w.width, 0]
    vec2 = np.array([vec1[1], vec1[0], 0])
    pos_new = [pos[0], pos[1], -4000] + vec2 * 3
    camera_opts = dict(
        pos=pos_new,
        focalPoint=pos,
        viewup=[0., 0., -1.]
    )

    show(
        meshes_vedo,
        w.world,
        lines,
        targets,
        arrows,
        boxes,
        axes,
        points,
        bg='lightgrey',
        bg2='lavender',
        camera=camera_opts,
        interactorStyle=10,
        resetcam=True,
        interactive=True,
        # verbose=True,
        title=f'welleng {VERSION}'
    )


def get_start_location(start_locations):
    start_location = np.average(start_locations, axis=0)
    start_location[2] = np.amin(start_locations[:, 2], axis=0)
    return start_location


def get_bb(vertices, min_size=[1000., 1000., 0.]):
    bb_max = np.amax(vertices.reshape(-1, 3), axis=0)
    bb_min = np.amin(vertices.reshape(-1, 3), axis=0)

    l, w, h = np.amax(np.vstack([(bb_max - bb_min), min_size]), axis=0)
    bb_center = bb_min + np.array(bb_max - bb_min) / 2

    world = World(
        bb_center,
        l,
        w,
        h
    )

    return world


# make a dictionary of axes options
def get_axes(world):
    assert VEDO, "ImportError: try pip install welleng[easy]"
    axes = Axes(
        world,
        xtitle='y: North (m)',  # swap axis to plot correctly
        ytitle='x: East (m)',
        ztitle='z: TVD (m)',
        xTitleJustify='bottom-right',
        yTitleJustify='top-right',
        zTitleJustify='top-right',
        xyGrid2=True, xyGrid=False,
        zxGrid=True, yzGrid2=True,
        zxGridTransparent=True, yzGrid2Transparent=True,
        yzGrid=False,
        xLabelRotation=-1,
        yLabelRotation=1,
        zLabelRotation=1,
    )

    for a in axes.unpack():  # unpack the Assembly to access its elements
        if 'title' in a.name or 'NumericLabel' in a.name:
            a.mirror('y')
        if 'yNumericLabel' in a.name:
            a.scale(0.8)
    return axes


def get_lines(clearance):
    """
    Add lines per reference well interval between the closest points on the
    reference well and the offset well and color them according to the
    calculated Separation Factor (SF) between the two wells at these points.

    Parameters
    ----------
        clearance: welleng.clearance object

    Returns
    -------
        lines: vedo.Lines object
            A vedo.Lines object colored by the object's SF values.
    """
    assert VEDO, "ImportError: try pip install welleng[easy]"
    c = clearance.SF
    start_points, end_points = clearance.get_lines()
    lines = Lines(start_points, end_points).cmap('hot_r', c, on='cells')
    lines.addScalarBar(title='SF')

    return lines


def figure(obj, type='scatter3d', **kwargs):
    assert PLOTLY, "ImportError: try pip install plotly"
    func = {
        'scatter3d': _scatter3d,
        'mesh3d': _mesh3d
    }
    fig = func[type](obj, **kwargs)
    return fig


def _scatter3d(survey, **kwargs):
    fig = go.Figure()
    fig.add_trace(
        go.Scatter3d(
            x=survey.e,
            y=survey.n,
            z=survey.tvd,
            name='survey',
            mode='lines',
            hoverinfo='skip'
        )
    )
    if not hasattr(survey, "interpolated"):
        survey.interpolated = [False] * len(survey.md)
    if survey.interpolated is None:
        survey.interpolated = [False] * len(survey.md)
    try:
        n, e, v, md, inc, azi = np.array([
            [n, e, v, md, inc, azi]
            for n, e, v, i, md, inc, azi in zip(
                survey.n, survey.e, survey.tvd, survey.interpolated,
                survey.md, survey.inc_deg, survey.azi_grid_deg
            )
            if bool(i) is True
        ]).T
        if n.size:
            text = [
                f"N: {n:.2f}m<br>E: {e:.2f}m<br>TVD: {v:.2f}m<br>"
                + f"MD: {md:.2f}m<br>INC: {inc:.2f}\xb0<br>AZI: {azi:.2f}\xb0"
                for n, e, v, md, inc, azi in zip(n, e, v, md, inc, azi)
            ]
            fig.add_trace(
                go.Scatter3d(
                    x=e,
                    y=n,
                    z=v,
                    name='interpolated',
                    mode='markers',
                    marker=dict(
                        size=5,
                        color='blue',
                    ),
                    text=text,
                    hoverinfo='text'
                )
            )
    except ValueError:
        pass

    try:
        n, e, v, md, inc, azi = np.array([
            [n, e, v, md, inc, azi]
            for n, e, v, i, md, inc, azi in zip(
                survey.n, survey.e, survey.tvd, survey.interpolated,
                survey.md, survey.inc_deg, survey.azi_grid_deg
            )
            if bool(i) is False
        ]).T
        text = [
            f"N: {n:.2f}m<br>E: {e:.2f}m<br>TVD: {v:.2f}m<br>"
            + f"MD: {md:.2f}m<br>INC: {inc:.2f}\xb0<br>AZI: {azi:.2f}\xb0"
            for n, e, v, md, inc, azi in zip(n, e, v, md, inc, azi)
        ]
        fig.add_trace(
            go.Scatter3d(
                x=e,
                y=n,
                z=v,
                name='survey_point',
                mode='markers',
                marker=dict(
                    size=5,
                    color='red',
                ),
                text=text,
                hoverinfo='text'
            )
        )
    except ValueError:
        pass

    fig = _update_fig(fig, kwargs)

    return fig


def _mesh3d(mesh, **kwargs):
    fig = go.Figure()
    vertices = mesh.vertices.reshape(-1, 3)
    faces = np.array(mesh.faces)
    n, e, v = vertices.T
    i, j, k = faces.T
    fig.add_trace(
        go.Mesh3d(
            x=e, y=n, z=v,
            i=i, j=j, k=k,
        )
    )
    if kwargs.get('edges'):
        tri_points = np.array([
            vertices[i] for i in faces.reshape(-1)
        ])
        n, e, v = tri_points.T
        fig.add_trace(
            go.Scatter3d(
                x=e, y=n, z=v,
                mode='lines',
            )
        )
    fig = _update_fig(fig, kwargs)

    return fig


def _update_fig(fig, kwargs):
    """
    Update the fig axis along with any user defined kwargs.
    """
    fig.update_scenes(
        zaxis_autorange="reversed",
        aspectmode='data',
        xaxis=dict(
            title='East (m)'
        ),
        yaxis=dict(
            title='North (m)',
        ),
        zaxis=dict(
            title="TVD (m)"
        )
    )
    for k, v in kwargs.items():
        if k == "layout":
            fig.update_layout(v)
        elif k == "traces":
            fig.update_traces(v)
        else:
            continue

    return fig



import numpy as np
from .utils import linear_convert

# This model is EXPERIMENTAL, I wrote it more as a proof of concept for
# well trajectory optimization. This model has not been validated against
# know good data and I've not cross-checked the formulas against those
# referenced in the Pro Well Plan module that I based this on.


class TorqueDrag:
    def __init__(
        self,
        survey,
        assembly,
        fluid,
        torque_on_bit=0.,
        weight_on_bit=0.,
        overpull=0.,
        fixed_depth=None,
        unit='metric'
    ):
        """
        Parameters
        ----------
            weight_on_bit: float
                If unit is 'metric' then in tonnes, else if 'imperial' then
                kips.
            torque_on_bit: float
                If unit is 'metric' then in kNm, else if 'imperial'
                then kft.lbs.
        """
        self.G = 9.81
        self.survey = survey
        self.assembly = assembly
        self.fluid = fluid
        self = get_wob(self, weight_on_bit, overpull, unit)
        self = get_tob(self, torque_on_bit, unit)

        # add BHA profiles
        self._get_buoyancy()
        self.assembly_weight = np.flip(np.full_like(
            self.survey.radius, self.assembly.weight_metric * self.G
        ) * self.survey.delta_md)
        # self.assembly_weight = (
        #     self.assembly.density_metric * 1000 * self.G * np.pi
        #     * self.assembly_area * np.flip(self.survey.delta_md)
        # )
        self.buoyancy_factor = (
            (self.assembly.density_metric - self.fluid.density_metric)
            / self.assembly.density_metric
        )
        self.assembly_bouyed_weight = (
            self.assembly_weight * self.buoyancy_factor
            # self.assembly_weight * self.buoyancy
        )

        self._get_delta_angles()
        self._get_coeffs()
        self._get_loads()

    def _get_loads(self):
        self.drag_rih = [self.wob_metric]
        self.torque_rih = [self.tob_metric]
        self.drag_neutral = [0.]
        self.torque_neutral = [0.]
        self.drag_pooh = [self.overpull_metric]
        self.torque_pooh = [0.]
        for i, params in enumerate(zip(
            self.delta_inc, self.delta_azi, self.inc_avg, self.A, self.B,
            self.C, self.friction_coeff, self.string_radius
        )):
            self._get_rih(params)
            self._get_neutral(params)
            self._get_pooh(params)

        self._cleanup_loads()

    def _get_delta_angles(self):
        self.delta_inc, self.delta_azi = (
            np.vstack(
                (
                    np.array([0., 0.]),
                    np.diff(
                        np.array([
                            self.survey.inc_rad,
                            self.survey.azi_grid_rad
                        ]).T[::-1], axis=0
                    )
                )
            )
        ).T
        self.inc_avg = np.flip(self.survey.inc_rad) - self.delta_inc / 2

    def _get_coeffs(self):
        # These coeffs are used in all the calcs... so just calculate them
        # once (efficiently with numpy) and serve them to the helper functions.
        self.A = self.delta_azi * np.sin(self.inc_avg)
        self.B = (
            self.assembly_bouyed_weight
            * np.sin(self.inc_avg)
        )
        self.C = (
            self.assembly_bouyed_weight
            * np.cos(self.inc_avg)
        )
        self.friction_coeff = np.flip(self.survey.friction_coeff)

        # TODO: evolve assembly into a string of components with a radius
        # profile
        self.string_radius = np.flip(np.full_like(
            self.survey.inc_rad,
            self.assembly.id_metric / (2 * 1000)  # convert to radius in meters
        ))

    def _get_rih(self, params):
        delta_inc, delta_azi, inc_avg, a, b, c, fc, r = params

        # calculate drag
        fn = (
            (self.drag_rih[-1] * a) ** 2
            + (self.drag_rih[-1] * delta_inc + b) ** 2
        ) ** 0.5

        delta_ft = (
            c - fc * fn
        )

        ft = self.drag_rih[-1] + delta_ft

        self.drag_rih.append(ft)

        # calculate torque
        delta_t = fc * fn * r

        t = self.torque_rih[-1] + delta_t

        self.torque_rih.append(t)

    def _get_neutral(self, params):
        delta_inc, delta_azi, inc_avg, a, b, c, fc, r = params

        # calculate drag
        fn = (
            (self.drag_neutral[-1] * a) ** 2
            + (self.drag_neutral[-1] * delta_inc + b) ** 2
        ) ** 0.5

        delta_ft = (
            c
        )

        ft = self.drag_neutral[-1] + delta_ft

        self.drag_neutral.append(ft)

        # calculate torque
        delta_t = fc * fn * r

        t = self.torque_neutral[-1] + delta_t

        self.torque_neutral.append(t)

    def _get_pooh(self, params):
        delta_inc, delta_azi, inc_avg, a, b, c, fc, r = params

        # calculate drag
        fn = (
            (self.drag_pooh[-1] * a) ** 2
            + (self.drag_pooh[-1] * delta_inc + b) ** 2
        ) ** 0.5

        delta_ft = (
            c + fc * fn
        )

        ft = self.drag_pooh[-1] + delta_ft

        self.drag_pooh.append(ft)

        # calculate torque
        delta_t = fc * fn * r

        t = self.torque_pooh[-1] + delta_t

        self.torque_pooh.append(t)

    def _cleanup_loads(self):
        loads = np.flip(np.array([
            self.drag_rih,
            self.torque_rih,
            self.drag_neutral,
            self.torque_neutral,
            self.drag_pooh,
            self.torque_pooh,
        ]) / 1000, axis=-1)

        (
            self.drag_rih,
            self.torque_rih,
            self.drag_neutral,
            self.torque_neutral,
            self.drag_pooh,
            self.torque_pooh,
        ) = loads

    def _get_buoyancy(self):
        self._get_areas()
        self.buoyancy = (
            1 - (self.fluid.density_metric * self.annulus_area)
            / (self.assembly.density_metric * (
                self.annulus_area - self.assembly_area
            ))
        )

    def _get_areas(self):
        self.annulus_area = np.flip(
            np.pi * (
                self.survey.radius ** 2
                - (self.assembly.od_metric / (2 * 1000)) ** 2
            )
        )
        self.assembly_area = np.flip(
            np.pi * (
                (self.assembly.od_metric / (2 * 1000)) ** 2
                - (self.assembly.id_metric / (2 * 1000)) ** 2
            )
        )


def get_wob(obj, wob, op, unit, factor=2.204622622):
    if unit == 'imperial':
        obj.wob_imperial = wob * 1000
        obj.overpull_imperial = op * 1000
        obj.wob_metric, obj.overpull_metric = linear_convert(
            [wob, op], 1/factor
        )
    else:
        obj.wob_metric = wob * 1000
        obj.overpull_metric = op * 1000
        obj.wob_imperial, obj.overpull_imperial = linear_convert(
            [wob, op], factor
        )

    return obj


def get_tob(obj, tob, unit, factor=1.36):
    if unit == 'imperial':
        obj.tob_imperial = tob * 1000
        obj.tob_metric = linear_convert(
            tob, factor
        )
    else:
        obj.tob_metric = tob * 1000
        obj.tob_imperial = linear_convert(
            tob, 1/factor
        )

    return obj
